import asyncio
from aiohttp import web, errors as weberrors
import socket
import threading
import atexit
import sys
import time
from queue import Queue, Empty

UDP_IN_PORT=6666
UDP_OUT_PORT=6668

received_logs = list()
log_store_limit = 200
log_resend_limit = 25

@asyncio.coroutine
def netconsole_monitor():
    # set up receiving socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind( ('',UDP_IN_PORT) )

    # set up atexit handler to close sockets
    def atexit_func():
        sock.close()

    atexit.register(atexit_func)

    # set up threads to emulate non-blocking io
    # thread-level emulation required for compatibility with windows
    sock_queue = Queue()

    def enqueue_output_sock(s, q):
        while True:
            q.put(s.recv(4096))

    sock_reader = threading.Thread(target=enqueue_output_sock, args=(sock, sock_queue))
    sock_reader.daemon = True
    sock_reader.start()

    # main loop
    while True:
        try:
            msg = str(sock_queue.get_nowait(), 'utf-8')

        except Empty:
            pass # no output
        else:
            # Send msg to the socket
            try:
                process_log(msg)
            except web.WSClientDisconnectedError as exc:
                print(exc.code, exc.message)
                return
        yield from asyncio.sleep(0.05)

def process_log(message):
    log_data = {"message": message, "timestamp": time.monotonic()}
    received_logs.append(log_data)
    while len(received_logs) > log_store_limit:
        received_logs.remove(received_logs[0])
    for websocket in websocket_connections[:]:
        try:
            websocket.send_str(message)
        except weberrors.ClientDisconnectedError or weberrors.WSClientDisconnectedError:
            pass


websocket_connections = list()

@asyncio.coroutine
def netconsole_websocket(request):

    # Start websocket
    websocket = web.WebSocketResponse()
    websocket.start(request)

    # Get websocket ID and save the object
    websocket_id = len(websocket_connections)
    websocket_connections.append(websocket)

    # Resend last collected logs
    start_id = min(len(received_logs), log_resend_limit)
    for log in received_logs[-start_id:]:
        websocket.send_str(log["message"])

    # Do other stuff
    print("NC Websocket {} Connected".format(websocket_id))
    yield from netconsole_websocket_keepalive(websocket)
    print("NC Websocket {} Disonnected".format(websocket_id))
    return websocket

@asyncio.coroutine
def netconsole_websocket_keepalive(ws):
    while True:
        try:
            data = yield from ws.receive_str()
        except Exception:
            websocket_connections.remove(ws)
            return
        print(data)

@asyncio.coroutine
def netconsole_log_dump(request):
    print("Dumping logs to request.")
    data = "\n".join(l["message"] for l in received_logs)
    return web.Response(body=data.encode('utf-8'))


# /usr/bin/python3
import asyncio
import sys
from os import symlink, remove
from os.path import abspath, dirname, join, exists

from aiohttp import web

from webdash import netconsole_controller
from webdash import networktables_controller

@asyncio.coroutine
def forward_request(request):
    return web.HTTPFound("/index.html")

INIT_FILE = "Webdash_init.sh"
INSTALL_LOCATIONS = "/etc/init.d/" + INIT_FILE, "/etc/rc5.d/S99" + INIT_FILE

def run_server(port):
    print("Starting Webdash Server.")
    file_root = join(abspath(dirname(__file__)), "resources")
    asyncio.async(netconsole_controller.netconsole_monitor())
    networktables_controller.setup_networktables()
    app = web.Application()
    app.router.add_route("GET", "/networktables", networktables_controller.networktables_websocket)
    app.router.add_route("GET", "/netconsole", netconsole_controller.netconsole_websocket)
    app.router.add_route("GET", "/netconsole_dump", netconsole_controller.netconsole_log_dump)
    app.router.add_route("GET", "/", forward_request)
    app.router.add_static("/", file_root)
    loop = asyncio.get_event_loop()
    f = loop.create_server(app.make_handler(), '0.0.0.0', port)
    srv = loop.run_until_complete(f)
    print("RoboRIO Webdash listening on", srv.sockets[0].getsockname())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

def main():
    if len(sys.argv) <= 1:
        run_server(5801)
    elif sys.argv[1] == "install-initfile":
        for tgt in INSTALL_LOCATIONS:
            if not exists(dirname(tgt)):
                print("ERROR: Installation path {} does not exist. (Are you sure you are "
                      "installing this on an os with sysvinit?)".format(dirname(tgt)))
                exit(-1)
            src = join(abspath(dirname(__file__)), INIT_FILE)
            if exists(tgt):
                res = input("{} already exists, Remove? (y/N)".format(tgt))
                if res.lower() == "y":
                    remove(tgt)
                else:
                    print("ERROR: Target already exists.")
                    exit(-1)
            print("Symlinking {} to {}.".format(src, tgt))
            symlink(src, tgt)
            print("Successfully installed {}".format(INIT_FILE))
        exit(1)
    elif sys.argv[1] == "remove-initfile":
        for tgt in INSTALL_LOCATIONS:
            if not exists(tgt):
                print("ERROR: Init file {} not installed.".format(tgt))
            else:
                remove(tgt)
                print("Successfully removed {}".format(tgt))
        exit(1)
    else:
        if len(sys.argv) > 1:
            print("Unknown option: {}".format(sys.argv[1]))
        print("Usage: webdash [ , install-initfile, remove-initfile]")
        exit(1)

if __name__ == "__main__":
    main()


from __future__ import print_function

from os.path import dirname, exists, join
import sys, subprocess
from setuptools import setup

setup_dir = dirname(__file__)
git_dir = join(setup_dir, '.git')
base_package = 'webdash'
version_file = join(setup_dir, base_package, 'version.py')

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(["git", "describe", "--tags", "--long", "--dirty=-dirty"],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode('utf-8').rstrip().split('-', 2)
    if commits != '0' or '-dirty' in local:
        v = '%s.post0.dev%s' % (v, commits)

    # Create the version.py file
    with open(version_file, 'w') as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))

with open(version_file, 'r') as fp:
    exec(fp.read(), globals())

with open(join(setup_dir, 'README.rst'), 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name='roborio-webdash',
    version=__version__,
    description='A web-based dashboard with several useful utilities',
    long_description=long_description,
    author='Christian Balcom',
    author_email='robot.inventor@gmail.com',
    keywords='frc first robotics networktables netconsole livewindow',
    packages=[base_package],
    include_package_data=True,
    install_requires=[
        "aiohttp",
        "pynetworktables"
    ],
    license='BSD License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License'
    ],
    entry_points={
        'console_scripts': [
            'webdash=webdash.main:main',
        ],
    }
)


from networktables import NetworkTable
import asyncio
import json
import math
from aiohttp import web, errors as weberrors
from threading import RLock
from copy import deepcopy

ip_address = "127.0.0.1"

initialized_networktables = False

table_data = dict()
table_data_lock = RLock()
root_table = None

connections = list()
tagged_tables = list()

class ConnectionListener:
    def connected(self, table):
        set_local_value("~CONNECTED~", True)

    def disconnected(self, table):
        set_local_value("~CONNECTED~", False)

def val_listener(key, value, isNew):
    #print("val_listener: key={}, value={}, type={}".format(key, value, type(value)))
    set_local_value(key, value)

def get_local_value(key):
    with table_data_lock:
        if key[0] == NetworkTable.PATH_SEPARATOR:
            key = key[1:]
        value = table_data
        for s in key.split(NetworkTable.PATH_SEPARATOR):
            if s not in value:
                return None
            value = value[s]
        return value

def set_local_value(key, value):
    with table_data_lock:
        if isinstance(value, float) and (math.isnan(value) or math.isinf(value)):
            value = 0.0
        if key[0] == NetworkTable.PATH_SEPARATOR:
            key = key[1:]
        keysplit = key.split(NetworkTable.PATH_SEPARATOR)
        value_key = keysplit[-1:][0]
        table_key = keysplit[:-1]
        target_table = table_data

        for s in table_key:
            if s not in target_table:
                target_table[s] = dict()
            target_table = target_table[s]

        # Save the value if it is new
        if value_key == "":
            return

        if value_key in target_table:
            value = type(target_table[value_key])(value)

        target_table[value_key] = value
        trigger_update()

def trigger_update():
    for con in connections:
        con["updated_data"] = True

def set_value(key, value):
    try:
        current_value = get_local_value(key)
        if current_value is not None:
            value = to_type(value, type(current_value))
        if key[0] == NetworkTable.PATH_SEPARATOR:
            key = key[1:]

        if isinstance(value, bool):
            root_table.putBoolean(key, value)
        elif isinstance(value, float) or isinstance(value, int):
            root_table.putNumber(key, value)
        else:
            root_table.putString(key, str(value))
    except Exception as e:
        print(e)
    finally:
        trigger_update()


def to_type(value, target_type):
    value = str(value)
    if target_type is bool:
        return value.lower() in ("yes", "true", "t", "1")
    elif target_type is int or target_type is float:
        return float(value)
    else:
        return value

def setup_networktables(ip=ip_address):
    global root_table, table_data, initialized_networktables
    if initialized_networktables:
        return
    NetworkTable.setIPAddress(ip)
    NetworkTable.setClientMode()
    NetworkTable.initialize()
    root_table = NetworkTable.getTable("")
    c_listener = ConnectionListener()
    root_table.addConnectionListener(c_listener)
    root_table.addGlobalListener(val_listener, True)
    initialized_networktables = True

@asyncio.coroutine
def networktables_websocket(request):
    # Setup websocket
    ws = web.WebSocketResponse()
    ws.start(request)

    # Setup connection dict
    con_id = len(connections)
    with table_data_lock:
        connection = {"socket": ws, "updated_data": True}
        connections.append(connection)
    print("NT Websocket {} Connected".format(con_id))

    # Start listener coroutine
    asyncio.async(networktables_websocket_listener(ws))

    # Set IP status data
    ip = request.transport.get_extra_info("sockname")[0]
    set_local_value("~SERVER_IP~", ip)

    last_data = dict()

    # Update periodically until the websocket is closed.
    try:
        while True:
            yield from asyncio.sleep(1)
            while True:
                yield from asyncio.sleep(.1)
                if connection["updated_data"]:
                    connection["updated_data"] = False
                    updates = dict_delta(last_data, table_data)
                    string_data = json.dumps(updates)
                    #print("Sending " + string_data)
                    ws.send_str(string_data)
                    last_data = deepcopy(table_data)
                if ws.closing:
                    break
            if ws.closing:
                break
    except weberrors.ClientDisconnectedError or weberrors.WSClientDisconnectedError:
        print("Client Disconnected")
    finally:
        print("NT Websocket {} Disconnected".format(con_id))
        with table_data_lock:
            connections.remove(connection)
        return ws

def dict_delta(dict_a, dict_b):
    """
    recursively compares two dictionaries, returns the dictionary of differences.
    aka retval = dict_b - dict_a
    """
    result = dict()
    for k in dict_b:
        if k in dict_a:
            if isinstance(dict_a[k], dict) and isinstance(dict_b[k], dict):
                comp_res = dict_delta(dict_a[k], dict_b[k])
                if len(comp_res) > 0:
                    result[k] = comp_res
            elif dict_a[k] != dict_b[k]:
                result[k] = dict_b[k]
        else:
            result[k] = dict_b[k]
    return result


@asyncio.coroutine
def networktables_websocket_listener(ws):
    while True:
        try:
            jdata = yield from ws.receive_str()
        except Exception:
            return
        data = json.loads(jdata)

        set_value(data["key"], data["value"])